# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""End-to-end tests for the plugins on an enterprise URL"""

import urllib
from pathlib import Path

import pytest
from dbrownell_Common.TestHelpers.StreamTestHelpers import (
    InitializeStreamCapabilities,
)
from typer.testing import CliRunner
from utilities import CheckPATFileExists, GetGithubUrl, ScrubDurationGithuburlAndSpaces

from RepoAuditor.EntryPoint import app

# ----------------------------------------------------------------------
pytest.fixture(InitializeStreamCapabilities(), scope="session", autouse=True)


# ----------------------------------------------------------------------
@pytest.fixture
def args() -> list[str]:
    return [
        "--include",
        "GitHub",
        "--GitHub-url",
        GetGithubUrl("enterprise"),
        # Currently GitHub Enterprise does not support GitHub hosted runners.
        # As a result, no status checks show up to select and this will always fail.
        "--GitHub-EnsureStatusChecks-disabled",
    ]


# ----------------------------------------------------------------------
@pytest.fixture
def pat_args(args) -> list[str]:
    github_pat_filename = (Path(__file__).parent / "github_gatech_pat.txt").resolve()
    parsed = urllib.parse.urlparse(GetGithubUrl("enterprise"))
    parsed = parsed._replace(path="")
    CheckPATFileExists(github_pat_filename, github_url=parsed.geturl())

    return args + ["--GitHub-pat", str(github_pat_filename)]


# ----------------------------------------------------------------------
def test_Successful(pat_args, snapshot):
    """Test if RepoAuditor runs successfully with default options."""
    result = CliRunner().invoke(app, pat_args)

    assert result.exit_code == 0, result.output
    assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
class TestStandard:
    # ----------------------------------------------------------------------
    def test_NoAutoMerge(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-AutoMerge-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoDeleteHeadBranches(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-DeleteHeadBranches-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    @pytest.mark.skip(reason="Not enabled by Enterprise Administrator")
    def test_NoDependabotSecurityUpdates(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-DependabotSecurityUpdates-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoMergeCommit(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-MergeCommit-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesRebaseMergeCommit(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-RebaseMergeCommit-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    @pytest.mark.skip(reason="Not enabled by Enterprise Administrator")
    def test_NoSecretScanning(self, pat_args, snapshot):
        """Test is secret scanning is not disabled."""
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SecretScanning-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    @pytest.mark.skip(reason="Not enabled by Enterprise Administrator")
    def test_NoSecretScanningPushProtection(self, pat_args, snapshot):
        """Test is secret scanning push protection is not disabled."""
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SecretScanningPushProtection-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesSquashCommitMerge(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SquashCommitMerge-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesSuggestUpdatingPullRequestBranches(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SuggestUpdatingPullRequestBranches-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesSupportDiscussions(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportDiscussions-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSupportIssues(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportIssues-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSupportProjects(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportProjects-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSupportWikis(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportWikis-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesTemplateRepository(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-TemplateRepository-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoWebCommitSignoff(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-WebCommitSignoff-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_DefaultBranchValue(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-DefaultBranch-value", "not_main"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_LicenseValue(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-License-value", "Not the MIT License"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_MergeCommitMessageValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--GitHub-MergeCommitMessage-value",
                "This is not a valid merge commit message",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_PrivateValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-Private-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_SquashMergeCommitMessageValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--GitHub-SquashMergeCommitMessage-value",
                "This is not a valid squash merge commit message",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_Protected(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-Protected-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
class TestClassic:
    # ----------------------------------------------------------------------
    def test_AllowDeletions(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-AllowDeletions-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_AllowForcePush(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-AllowMainlineForcePushes-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_DismissStalePullRequestApprovals(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-DismissStalePullRequestApprovals-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_DoNotAllowBypassSettings(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-DoNotAllowBypassSettings-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    @pytest.mark.skip(reason="GitHub hosted runners not available on Enterprise")
    def test_EnsureStatusChecks(self, pat_args):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-EnsureStatusChecks-disable", "--verbose"],
        )

        assert result.exit_code == 0, result.output
        assert "[DoesNotApply] EnsureStatusChecks" in result.stdout

    # ----------------------------------------------------------------------
    def test_RequireApprovalMostRecentPush(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireApprovalMostRecentPush-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireApprovals(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireApprovals-value", "2"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireCodeOwnerReview(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireCodeOwnerReview-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireConversationResolution(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireConversationResolution-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireLinearHistory(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireLinearHistory-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequirePullRequests(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequirePullRequests-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireSignedCommits(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireSignedCommits-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireStatusChecksToPass(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireStatusChecksToPass-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireUpToDateBranches(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireUpToDateBranches-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the CommunityStandardsModule object."""

from RepoAuditor.Module import ExecutionStyle
from RepoAuditor.Plugins.CommunityStandards.CommunityStandardsQuery import CommunityStandardsQuery
from RepoAuditor.Plugins.GitHubBase.Module import GitHubBaseModule


class CommunityStandardsModule(GitHubBaseModule):
    """Module that validates existence of repository files for Community Standards."""

    def __init__(self) -> None:
        super().__init__(
            "CommunityStandards",
            "Validates existence of repository files for Community Standards.",
            ExecutionStyle.Parallel,
            [
                CommunityStandardsQuery(),
            ],
            requires_explicit_include=True,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireLinearHistory object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireLinearHistory(ClassicEnableRequirementImpl):
    """Require a linear history."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireLinearHistory",
            False,
            "enabled",
            "Protect matching branches",
            "Require linear history",
            lambda data: data["branch_protection_data"]["required_linear_history"]["enabled"],
            textwrap.dedent(
                """\
                The default behavior is to not require a linear history as this option is disabled when rebase
                merging and squash merging are disabled (which are the default validation settings).

                Reasons for this Default
                ------------------------
                - This option is disabled within the GitHub UX when rebase merging and squash merging are disabled
                  (which are the default validation settings).

                Reasons to Override this Default
                --------------------------------
                - You have enabled rebase merging or squash merging
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the ClassicBranchProtection object.

The API request/response has the following schema:

REQUEST : https://api.github.com/repos/<username>/<repo>/branches/<branch>/protection
RESPONSE:
{
    "url": <not used>,
    "required_status_checks": { /* RequireStatusChecksToPass */
        "url": <not used>,
        "strict": RequireUpToDateBranches,
        "contexts": <not used>,
        "contexts_url": <not used>,
        "checks": EnsureStatusChecks
    },
    "required_pull_request_reviews": {  /* RequirePullRequests */
        "url": <not used>,
        "dismiss_stale_reviews": DismissStalePullRequestApprovals,
        "require_code_owner_reviews": RequireCodeOwnerReview,
        "require_last_push_approval": RequireApprovalMostRecentPush,
        "required_approving_review_count": RequireApprovals,
        "bypass_pull_request_allowances": <not used>,
    },
    "required_signatures": {
        "url": <not used>,
        "enabled": RequireSignedCommits
    },
    "enforce_admins": {
        "url": <not used>,
        "enabled": DoNotAllowBypassSettings
    },
    "required_linear_history": {
        "enabled": RequireLinearHistory
    },
    "allow_force_pushes": {
        "enabled": AllowMainlineForcePushes
    },
    "allow_deletions": {
        "enabled": AllowDeletions
    },
    "block_creations": {
        "enabled": <not used>
    },
    "required_conversation_resolution": {
        "enabled": RequireConversationResolution
    },
    "lock_branch": {
        "enabled": <not used>
    },
    "allow_fork_syncing": {
        "enabled": <not used>
    }
}
"""

from typing import Any, Optional

import requests
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.AllowDeletions import AllowDeletions
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.AllowMainlineForcePushes import (
    AllowMainlineForcePushes,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.DismissStalePullRequestApprovals import (
    DismissStalePullRequestApprovals,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.DoNotAllowBypassSettings import (
    DoNotAllowBypassSettings,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.EnsureStatusChecks import (
    EnsureStatusChecks,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireApprovalMostRecentPush import (
    RequireApprovalMostRecentPush,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireApprovals import RequireApprovals
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireCodeOwnerReview import (
    RequireCodeOwnerReview,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireConversationResolution import (
    RequireConversationResolution,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireLinearHistory import (
    RequireLinearHistory,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequirePullRequests import (
    RequirePullRequests,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireSignedCommits import (
    RequireSignedCommits,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireStatusChecksToPass import (
    RequireStatusChecksToPass,
)
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireUpToDateBranches import (
    RequireUpToDateBranches,
)
from RepoAuditor.Query import ExecutionStyle, Query


# ----------------------------------------------------------------------
class ClassicBranchProtectionQuery(Query):
    """Query with requirements that operate on class branch protection rules."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "ClassicBranchProtectionQuery",
            ExecutionStyle.Parallel,
            [
                RequirePullRequests(),
                RequireApprovals(),
                DismissStalePullRequestApprovals(),
                RequireCodeOwnerReview(),
                RequireApprovalMostRecentPush(),
                RequireStatusChecksToPass(),
                RequireUpToDateBranches(),
                EnsureStatusChecks(),
                RequireConversationResolution(),
                RequireSignedCommits(),
                RequireLinearHistory(),
                DoNotAllowBypassSettings(),
                AllowDeletions(),
                AllowMainlineForcePushes(),
            ],
        )

    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Get the data from an API session."""
        branch = module_data.get("branch")
        if branch is None:
            # Get the default branch name
            response = module_data["session"].get("")

            response.raise_for_status()
            response = response.json()

            branch = response["default_branch"]

        module_data["branch"] = branch

        # Get the information for the default branch
        response = module_data["session"].get(f"branches/{module_data['branch']}")

        response.raise_for_status()
        response = response.json()

        module_data["branch_data"] = response

        if not module_data["branch_data"].get("protected", False):
            return None

        # Note that once here, we know that the branch is protected, but we don't know the
        # protection scheme used (rule sets or classic). Attempt to get the classic information
        # and then see if rule sets are in use if the classic information is not found.
        response = module_data["session"].get(f"/branches/{module_data['branch']}/protection")

        if response.status_code == requests.codes.NOT_FOUND:
            # Does this branch use rule sets?
            ruleset_response = module_data["session"].get(f"rules/branches/{module_data['branch']}")

            ruleset_response.raise_for_status()
            ruleset_response = ruleset_response.json()

            # If there is data, assume that the branch is protected by rule sets
            if ruleset_response:
                return None

            # Classic branch protection information is only accessible when a pat is provided. Exit
            # gracefully if there isn't a PAT, as the DefaultBranchQuery will print a warning if
            # the PAT wasn't provided.
            if module_data["session"].github_pat is None:
                return None

            # If here, let the error result in an exception in the code that follows.

        response.raise_for_status()
        response = response.json()

        module_data["branch_protection_data"] = response

        return module_data



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for GitHub/Module.py"""

from pathlib import Path

from RepoAuditor.Plugins.GitHub.Module import GitHubModule
from RepoAuditor.Plugins.GitHubBase.Module import _GitHubSession


class TestGitHubModule:
    """Unit tests for the Module class in the GitHub plugin."""

    def test_Construct(self):
        """Test constructor."""
        module = GitHubModule()
        assert isinstance(module, GitHubModule)

    def test_GenerateInitialData(self):
        """Test GenerateInitialData method."""
        dynamic_args = {
            "url": "https://github.com/gt-sse-center/RepoAuditor",
            "pat": Path(__file__).parent / "dummy_github_pat.txt",
        }
        module = GitHubModule()
        dynamic_args = module.GenerateInitialData(dynamic_args)

        assert "session" in dynamic_args
        assert isinstance(dynamic_args["session"], _GitHubSession)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Common fixtures for GitHubModule unit tests."""

import pytest


@pytest.fixture
def session():
    """Create a dummy GitHub API session object."""

    class _GithubSession_:
        pass

    s = _GithubSession_()
    s.github_url = "https://github.com/gt-sse-center/RepoAuditor"
    s.pat = "github_pat_dummy"
    s.is_enterprise = False
    return s



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireCodeOwnerReview object."""

import textwrap
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireCodeOwnerReview(ClassicEnableRequirementImpl):
    """Requirement of review from code owners."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireCodeOwnerReview",
            False,
            "enabled",
            "Protect matching branches",
            "Require a pull request before merging -> Require review from Code Owners",
            _GetValue,
            textwrap.dedent(
                """\
                The default behavior is to not require explicit approval from a Code Owner.

                Reasons for this Default
                ------------------------
                - Most repositories are not large enough to warrant specialized reviews by people with specific knowledge about different
                  areas of the code.

                Reasons to Override this Default
                --------------------------------
                - You have a large repository with specialized areas of code that require reviews from specific people.

                * Note: consider reducing the barrier to entry for new contributors by simplifying the design and/or adding automated tests that
                  catch common problems before they are introduced into the code base.
                """,
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(
    data: dict[str, Any],
) -> Optional[bool]:
    settings = data["branch_protection_data"].get("required_pull_request_reviews", None)
    if settings is None:
        return None

    return settings.get("require_code_owner_reviews", None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequirePullRequests object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequirePullRequests(ClassicEnableRequirementImpl):
    """Require a pull request before merging."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequirePullRequests",
            True,
            "disabled",
            "Protect matching branches",
            "Require a pull request before merging",
            lambda data: "required_pull_request_reviews" in data["branch_protection_data"],
            textwrap.dedent(
                """\
                The default behavior is to require pull requests before merging.

                Reasons for this Default
                ------------------------
                - Pull requests are an important part of the development process and prevent unwanted changes from
                  making it into the mainline branch.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains common functionality that is used across different components."""

from RepoAuditor.Requirement import EvaluateResult, Requirement


# ----------------------------------------------------------------------
def CreateIncompleteDataResult() -> Requirement.EvaluateImplResult:
    """Create an incomplete data result."""
    return Requirement.EvaluateImplResult(
        EvaluateResult.Warning,
        "Incomplete data was encountered; please provide the GitHub PAT or update the PAT's permissions if one was provided.",
    )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the StandardQuery object.

The API request/response has the following schema:

REQUEST: https://api.github.com/repos/<username>/<repo>
RESPONSE:
{
    "id": <not used>,
    "node_id": <not used>,
    "name": <not used>",
    "full_name": <not used>,
    "private": Private,
    "owner": {
        "login": <not used>,
        "id": <not used>,
        "node_id": <not used>,
        "avatar_url": <not used>,
        "gravatar_id": <not used>,
        "url": <not used>,
        "html_url": <not used>,
        "followers_url": <not used>,
        "following_url": <not used>,
        "gists_url": <not used>,
        "starred_url": <not used>,
        "subscriptions_url": <not used>,
        "organizations_url": <not used>,
        "repos_url": <not used>,
        "events_url": <not used>,
        "received_events_url": <not used>,
        "type": <not used>,
        "site_admin": <not used>
    },
    "html_url": <not used>,
    "description": Description,
    "fork": <not used>,
    "url": <not used>,
    "forks_url": <not used>,
    "keys_url": <not used>,
    "collaborators_url": <not used>,
    "teams_url": <not used>,
    "hooks_url": <not used>,
    "issue_events_url": <not used>,
    "events_url": <not used>,
    "assignees_url": <not used>,
    "branches_url": <not used>,
    "tags_url": <not used>,
    "blobs_url": <not used>,
    "git_tags_url": <not used>,
    "git_refs_url": <not used>,
    "trees_url": <not used>,
    "statuses_url": <not used>,
    "languages_url": <not used>,
    "stargazers_url": <not used>,
    "contributors_url": <not used>,
    "subscribers_url": <not used>,
    "subscription_url": <not used>,
    "commits_url": <not used>,
    "git_commits_url": <not used>,
    "comments_url": <not used>,
    "issue_comment_url": <not used>,
    "contents_url": <not used>,
    "compare_url": <not used>,
    "merges_url": <not used>,
    "archive_url": <not used>,
    "downloads_url": <not used>,
    "issues_url": <not used>,
    "pulls_url": <not used>,
    "milestones_url": <not used>,
    "notifications_url": <not used>,
    "labels_url": <not used>,
    "releases_url": <not used>,
    "deployments_url": <not used>,
    "created_at": <not used>,
    "updated_at": <not used>,
    "pushed_at": <not used>,
    "git_url": <not used>,
    "ssh_url": <not used>,
    "clone_url": <not used>,
    "svn_url": <not used>,
    "homepage": <not used>,
    "size": <not used>,
    "stargazers_count": <not used>,
    "watchers_count": <not used>,
    "language": <not used>,
    "has_issues": SupportIssues,
    "has_projects": SupportProjects,
    "has_downloads": <not used> (No way to set value in GitHub settings),
    "has_wiki": SupportWikis,
    "has_pages": <not used>,
    "has_discussions": SupportDiscussions,
    "forks_count": <not used>,
    "mirror_url": <not used>,
    "archived": <not used>,
    "disabled": <not used>,
    "open_issues_count": <not used>,
    "license": {
        "key": <not used>,
        "name": License,
        "spdx_id": <not used>,
        "url": <not used>,
        "node_id": <not used>
    },
    "allow_forking": <not used> (No way to set value in GitHub settings),
    "is_template": <not used>,
    "web_commit_signoff_required": WebCommitSignoff,
    "topics": <not used>,
    "visibility": Private,
    "forks": <not used>,
    "open_issues": <not used>,
    "watchers": <not used>,
    "default_branch": DefaultBranch,
    "permissions": {
        "admin": <not used>,
        "maintain": <not used>,
        "push": <not used>,
        "triage": <not used>,
        "pull": <not used>
    },
    "temp_clone_token": <not used>,
    "allow_squash_merge": SquashMergeCommit,
    "allow_merge_commit": MergeCommit,
    "allow_rebase_merge": RebaseMergeCommit,
    "allow_auto_merge": AutoMerge,
    "delete_branch_on_merge": DeleteHeadBranches,
    "allow_update_branch": SuggestUpdatingPullRequestBranches,
    "use_squash_pr_title_as_default": <not used> (No way to set value in GitHub settings),
    "squash_merge_commit_message": SquashMergeCommitMessage,
    "squash_merge_commit_title": <not used> (No way to set value in GitHub settings),
    "merge_commit_message": MergeCommitMessage,
    "merge_commit_title": <not used> (No way to set value in GitHub settings),
    "custom_properties": <not used>,
    "organization": {
        "login": <not used>,
        "id": <not used>,
        "node_id": <not used>,
        "avatar_url": <not used>,
        "gravatar_id": <not used>,
        "url": <not used>,
        "html_url": <not used>,
        "followers_url": <not used>,
        "following_url": <not used>,
        "gists_url": <not used>,
        "starred_url": <not used>,
        "subscriptions_url": <not used>,
        "organizations_url": <not used>,
        "repos_url": <not used>,
        "events_url": <not used>,
        "received_events_url": <not used>,
        "type": <not used>,
        "site_admin": <not used>
    },
    "security_and_analysis": {
        "secret_scanning": SecretScanning,
        "secret_scanning_push_protection": SecretScanningPushProtection,
        "dependabot_security_updates": DependabotSecurityUpdates,
        "secret_scanning_validity_checks": <not used> (The functionality doesn't seem to be available in the UX yet (https://github.blog/2023-10-04-introducing-secret-scanning-validity-checks-for-major-cloud-services/))
    },
    "network_count": <not used>,
    "subscribers_count": <not used>
}
"""

from typing import Any, Optional

from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.StandardRequirements.AutoMerge import AutoMerge
from RepoAuditor.Plugins.GitHub.StandardRequirements.DefaultBranch import DefaultBranch
from RepoAuditor.Plugins.GitHub.StandardRequirements.DeleteHeadBranches import DeleteHeadBranches
from RepoAuditor.Plugins.GitHub.StandardRequirements.DependabotSecurityUpdates import (
    DependabotSecurityUpdates,
)
from RepoAuditor.Plugins.GitHub.StandardRequirements.Description import Description
from RepoAuditor.Plugins.GitHub.StandardRequirements.License import License
from RepoAuditor.Plugins.GitHub.StandardRequirements.MergeCommit import MergeCommit
from RepoAuditor.Plugins.GitHub.StandardRequirements.MergeCommitMessage import MergeCommitMessage
from RepoAuditor.Plugins.GitHub.StandardRequirements.Private import Private
from RepoAuditor.Plugins.GitHub.StandardRequirements.RebaseMergeCommit import RebaseMergeCommit
from RepoAuditor.Plugins.GitHub.StandardRequirements.SecretScanning import SecretScanning
from RepoAuditor.Plugins.GitHub.StandardRequirements.SecretScanningPushProtection import (
    SecretScanningPushProtection,
)
from RepoAuditor.Plugins.GitHub.StandardRequirements.SquashCommitMerge import SquashCommitMerge
from RepoAuditor.Plugins.GitHub.StandardRequirements.SquashMergeCommitMessage import (
    SquashMergeCommitMessage,
)
from RepoAuditor.Plugins.GitHub.StandardRequirements.SuggestUpdatingPullRequestBranches import (
    SuggestUpdatingPullRequestBranches,
)
from RepoAuditor.Plugins.GitHub.StandardRequirements.SupportDiscussions import SupportDiscussions
from RepoAuditor.Plugins.GitHub.StandardRequirements.SupportIssues import SupportIssues
from RepoAuditor.Plugins.GitHub.StandardRequirements.SupportProjects import SupportProjects
from RepoAuditor.Plugins.GitHub.StandardRequirements.SupportWikis import SupportWikis
from RepoAuditor.Plugins.GitHub.StandardRequirements.TemplateRepository import TemplateRepository
from RepoAuditor.Plugins.GitHub.StandardRequirements.WebCommitSignoff import WebCommitSignoff
from RepoAuditor.Query import ExecutionStyle, Query


# ----------------------------------------------------------------------
class StandardQuery(Query):
    """Query with requirements that operate on basic GitHub repository data."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "StandardQuery",
            ExecutionStyle.Parallel,
            [
                # metadata
                Description(),
                License(),
                # settings page
                TemplateRepository(),
                WebCommitSignoff(),
                DefaultBranch(),
                # settings page (Features)
                SupportWikis(),
                SupportIssues(),
                SupportDiscussions(),
                SupportProjects(),
                # settings page (Pull Requests)
                MergeCommit(),
                MergeCommitMessage(),
                SquashCommitMerge(),
                SquashMergeCommitMessage(),
                RebaseMergeCommit(),
                SuggestUpdatingPullRequestBranches(),
                AutoMerge(),
                DeleteHeadBranches(),
                # settings page (Danger Zone)
                Private(),
                # settings/security_analysis
                DependabotSecurityUpdates(),
                SecretScanning(),
                SecretScanningPushProtection(),
            ],
        )

    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Get the data from an API session."""
        response = module_data["session"].get("")

        response.raise_for_status()
        response = response.json()

        module_data["standard"] = response

        return module_data



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the ExistsRequirementImpl object."""

from collections.abc import Sequence
from pathlib import Path
from typing import Any

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
class ExistsRequirementImpl(Requirement):
    """Requirement which implements check if a file exists."""

    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        filename: str,
        possible_locations: Sequence[str],
        resolution: str,
        rationale: str,
        *,
        dynamic_arg_name: str = "unrequired",
        requires_explicit_include: bool = False,
    ) -> None:
        """Construct.

        Args:
            name (str): The name of the requirement. Used for unique identification.
            filename (str): The name of the file in the git repository being checked for.
            possible_locations (Sequence[str]): List of potential paths in the repo to check if `filename` exists.
            resolution (str): Message on how to resolve the requirement in case of an error.
            rationale (str): Rationale message on why this requirement is needed.
            dynamic_arg_name (str, optional): Name of the runtime argument (e.g. from command line) to this requirement.Defaults to "unrequired", which if enabled causes the requirement to be skipped.
            requires_explicit_include (bool, optional): Flag checking if this requirement needs to be explicitly included in the invocation. Defaults to False.

        """
        super().__init__(
            name,
            f"Validates that {filename} file exists.",
            ExecutionStyle.Parallel,
            resolution,
            rationale,
            requires_explicit_include=requires_explicit_include,
        )

        self.dynamic_arg_name = dynamic_arg_name
        self.filename = filename

        self.possible_locations = possible_locations

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            self.dynamic_arg_name: (
                bool,
                typer.Option(
                    False,
                    help=f"Disable requirement that the file {self.filename} exists.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        # Check if `dynamic_arg_name` is `not unrequired`.
        if not requirement_args[self.dynamic_arg_name]:
            for location in self.possible_locations:
                # Get full file path in temporary repo directory and check if it exists
                # This checks both upper case and lower case variants
                full_file_path = Path(query_data["repo_dir"].name) / Path(location)

                if full_file_path.exists():
                    # If full_file_path is a directory, check if it isn't empty
                    if full_file_path.is_dir() and any(full_file_path.iterdir()):
                        return Requirement.EvaluateImplResult(
                            EvaluateResult.Success,
                            f"File found in {location} directory of the repository",
                        )

                    return Requirement.EvaluateImplResult(
                        EvaluateResult.Success,
                        f"{self.filename} found in repository",
                    )

            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                f"No {self.filename} file found.",
                provide_resolution=True,
                provide_rationale=True,
            )

        # Requirement flag not set so DoesNotApply
        return Requirement.EvaluateImplResult(EvaluateResult.DoesNotApply, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the AllowBypassSettings object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class DoNotAllowBypassSettings(ClassicEnableRequirementImpl):
    """Do not allow bypassing settings."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "DoNotAllowBypassSettings",
            True,
            "disabled",
            "Protect matching branches",
            "Do not allow bypassing the above settings",
            lambda data: data["branch_protection_data"]["enforce_admins"]["enabled"],
            textwrap.dedent(
                """\
                The default behavior is to not allow administrators to bypass branch protection settings.

                Reasons for this Default
                ------------------------
                - Ensure that all pull requests go through the same verification process.

                Reasons to Override this Default
                --------------------------------
                - The steps invoked during the verification process...
                    * ...are unreliable.
                    * ...take an excessive amount of time to complete.

                * Note that all of the reasons in this section are workarounds to address the underlying instability
                  of the steps invoked during the verification process. The ideal solution is to address the
                  underlying instability.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for CommunityStandardsQuery.py"""

import pytest
from git import Repo

from RepoAuditor.Plugins.CommunityStandards.CommunityStandardsQuery import (
    CommunityStandardsQuery,
    TemporaryDirectory,
)


class MockTemporaryDirectory:
    """A mock class to replace tempfile.TemporaryDirectory."""

    def __init__(self):
        self.name = "test_temp_directory"

    def cleanup(self):
        """Mocked cleanup method"""


@pytest.fixture(autouse=True)
def patch_temp_directory(monkeypatch):
    monkeypatch.delattr("tempfile.TemporaryDirectory")


class TestCommunityStandardsQuery:
    def test_GetData(self, module_data, monkeypatch):
        """Test the GetData method."""
        monkeypatch.setattr(
            TemporaryDirectory,
            "__init__",
            MockTemporaryDirectory.__init__,
        )

        def mock_clone_from(github_url, repo_dirname, branch="main"):
            """A mocked clone_from method for the git.Repo class."""
            pass

        monkeypatch.setattr(
            Repo,
            "clone_from",
            mock_clone_from,
        )

        query = CommunityStandardsQuery()
        query_data = query.GetData(module_data)

        assert query_data["repo_dir"].name == "test_temp_directory"

    def test_GetData_adds_pat(self, module_data, monkeypatch):
        """Test the GetData method adds PAT to URL"""
        github_pat = "pat-123"
        github_url = "https://github.com/owner/repo"
        expected_url = "https://pat-123@github.com/owner/repo"
        module_data["url"] = github_url
        module_data["pat"] = github_pat

        monkeypatch.setattr(
            TemporaryDirectory,
            "__init__",
            MockTemporaryDirectory.__init__,
        )

        def mock_clone_from(github_url, repo_dirname, branch="main"):
            assert github_url == expected_url

        monkeypatch.setattr(Repo, "clone_from", mock_clone_from)

        query = CommunityStandardsQuery()
        _ = query.GetData(module_data)

    def test_GetData_does_not_add_pat(self, module_data, monkeypatch):
        """Test the GetData method preserves the URL when no PAT is provided"""
        github_url = "https://github.com/owner/repo"
        expected_url = "https://github.com/owner/repo"
        module_data["url"] = github_url

        monkeypatch.setattr(
            TemporaryDirectory,
            "__init__",
            MockTemporaryDirectory.__init__,
        )

        def mock_clone_from(github_url, repo_dirname, branch="main"):
            assert github_url == expected_url

        monkeypatch.setattr(Repo, "clone_from", mock_clone_from)

        query = CommunityStandardsQuery()
        _ = query.GetData(module_data)

    def test_GetData_handles_gitpython_errors(self, module_data, monkeypatch):
        """Test the GetData method handles errors raised by the gitpyhon module"""
        gitpython_error_msg = "error cloning repository"

        monkeypatch.setattr(
            TemporaryDirectory,
            "__init__",
            MockTemporaryDirectory.__init__,
        )

        def mock_clone_from(github_url, repo_dirname, branch="main"):
            raise Exception(gitpython_error_msg)

        monkeypatch.setattr(Repo, "clone_from", mock_clone_from)

        query = CommunityStandardsQuery()
        with pytest.raises(RuntimeError) as e_info:
            _ = query.GetData(module_data)
        assert gitpython_error_msg in str(e_info)

    def test_Cleanup(self, module_data):
        """Test the Cleanup method."""
        query = CommunityStandardsQuery()
        module_data["repo_dir"] = MockTemporaryDirectory()
        result = query.Cleanup(module_data)

        assert result is None



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the TemplateRepository object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class TemplateRepository(StandardEnableRequirementImpl):
    """Template repository."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "TemplateRepository",
            False,
            "enabled",
            "settings",
            "General",
            "Template repository",
            lambda data: data["standard"].get("is_template", None),
            rationale=textwrap.dedent(
                """\
                The default behavior is that this is not a template repository.

                Reasons for this Default
                ------------------------
                - Most repositories are not templates.

                Reasons to Override this Default
                --------------------------------
                - Your repository is a template repository.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Helper utilities for Github tests"""

import re
import sys
import textwrap
from pathlib import Path
from typing import Match

from dbrownell_Common.TestHelpers.StreamTestHelpers import ScrubDuration
from git import Repo


# ----------------------------------------------------------------------
def ScrubSpaces(content: str) -> str:
    """
    Function to remove variable spaces after a match with a Github URL from a string.
    """

    def replace_func(match: Match) -> str:
        return match.group("scrubbed_github_url") + "<scrubbed-space>"

    return re.sub(
        r"(?P<scrubbed_github_url>'<scrubbed-github-url>')(\s+)",
        replace_func,
        content,
    )


# ----------------------------------------------------------------------
def ScrubGithubUrl(content: str) -> str:
    """
    Function to replace the GitHub username and (possibly) alternative repository name with the default one.
    """

    def replace_func(_: Match) -> str:
        return "'<scrubbed-github-url>'"

    return re.sub(
        r"('https?:\/\/.+')",
        replace_func,
        content,
    )


def ScrubDurationGithuburlAndSpaces(content: str) -> str:
    """Scrub the duration, GitHub urls, and variable spaces from the content."""
    scrubbed_duration = ScrubDuration(content)
    replaced_github_url = ScrubGithubUrl(scrubbed_duration)
    scrubbed_spaces = ScrubSpaces(replaced_github_url)
    return scrubbed_spaces


def GetGithubUrl(remote_name: str = "origin") -> str:
    """Get the Github username, repository and URL from the current git repository."""
    repo = Repo(__file__, search_parent_directories=True)

    try:
        # Get the URL for the remote specified by `remote_name`.
        repo_url = getattr(repo.remotes, remote_name).url
    except AttributeError as exc:
        # The `remote_name` does not exist, so we use a default.
        # Specifically for main repo CI.
        if remote_name == "origin":
            repo_url = "git@github.com:gt-sse-center/RepoAuditor.git"
        elif remote_name == "enterprise":
            repo_url = "git@github.gatech.edu:sse-center/RepoAuditor.git"
        else:
            # Invalid remote name so we throw exception
            raise exc

    if repo_url.startswith("git@"):
        # If the URL starts with 'git@', it is an SSH URL.
        # Convert it to HTTPS format.
        repo_url = repo_url.replace(":", "/").replace("git@", "https://")

    return repo_url.split(".git")[0]


def CheckPATFileExists(github_pat_filename: Path, github_url: str = "https://github.com"):
    """Check if PAT file exists, and give error message if not."""
    if not github_pat_filename.is_file():
        sys.stdout.write(
            textwrap.dedent(
                f"""\


            The filename '{github_pat_filename}' does not exist. Please create this file and add your GitHub Personal Access Token (PAT) to it.
            This git repository is configured to ignore the file so that it will never be included as part of a commit.

            To create a new token:

                1. Visit {github_url}/settings/tokens/new
                2. Ensure that 'repo' scope is checked
                3. Click 'Generate token'
                4. Copy the token to the clipboard
                5. Create the file '{github_pat_filename}'
                6. Paste the token into the created file
                7. Save the file and run these tests again.

            These tests query a repository on GitHub, but GitHub limits the number of concurrent requests made to a repository when a PAT is
            not provided. As a result, the tests will fail once the limit is reached.
            """,
            ),
        )

        sys.exit(-1)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""End-to-end tests for the CommunityStandards plugin."""

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest
from dbrownell_Common.TestHelpers.StreamTestHelpers import (
    InitializeStreamCapabilities,
)
from git import Repo
from typer.testing import CliRunner
from utilities import CheckPATFileExists, GetGithubUrl, ScrubDurationGithuburlAndSpaces

from RepoAuditor.EntryPoint import app
from RepoAuditor.Plugins.CommunityStandards.CommunityStandardsQuery import CommunityStandardsQuery

# ----------------------------------------------------------------------
pytest.fixture(InitializeStreamCapabilities(), scope="session", autouse=True)


def GetDataFunctional(path_to_delete: str, is_directory=False):
    """A functional which returns a modified GetData method which
    deletes the specified path in the cloned repo.
    """

    def GetData(self, module_data):
        # Clone the GitHub repository to a temp directory
        github_url = module_data["url"]
        branch = module_data.get("branch", "main")
        temp_repo_dir = TemporaryDirectory()
        Repo.clone_from(github_url, temp_repo_dir.name, branch=branch)

        path = Path(temp_repo_dir.name) / path_to_delete

        if is_directory:
            shutil.rmtree(path)
        else:
            path.unlink()

        # Record the path and the temp directory for later cleanup
        module_data["repo_dir"] = temp_repo_dir
        return module_data

    return GetData


# ----------------------------------------------------------------------
class TestCommunityStandards:
    """End-to-end tests for community standards files in a GitHub repository."""

    # ----------------------------------------------------------------------
    def test_CodeOfConduct(self, pat_args, snapshot):
        """Test if a CODE_OF_CONDUCT file exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == 0, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    @patch.object(
        CommunityStandardsQuery,
        "GetData",
        GetDataFunctional("CODE_OF_CONDUCT.md"),
    )
    def test_NoCodeOfConduct(self, pat_args, snapshot):
        """Test if a CODE_OF_CONDUCT file does not exist in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_CodeOwners(self, pat_args, snapshot):
        """Test if a CODEOWNERS file exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == 0, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    @patch.object(
        CommunityStandardsQuery,
        "GetData",
        GetDataFunctional(".github/CODEOWNERS"),
    )
    def test_NoCodeOwners(self, pat_args, snapshot):
        """Test if a CODEOWNERS file does not exist in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_Contributing(self, pat_args, snapshot):
        """Test if a CONTRIBUTING.md file exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == 0, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    @patch.object(
        CommunityStandardsQuery,
        "GetData",
        GetDataFunctional("CONTRIBUTING.md"),
    )
    def test_NoContributing(self, pat_args, snapshot):
        """Test if a CONTRIBUTING file does not exist in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_IssueTemplate(self, pat_args, snapshot):
        """Test if an Issue template file (e.g. ISSUE_TEMPLATE.md) exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )
        assert result.exit_code == 0, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    @patch.object(
        CommunityStandardsQuery,
        "GetData",
        GetDataFunctional(".github/ISSUE_TEMPLATE", True),
    )
    def test_NoIssueTemplate(self, pat_args, snapshot):
        """Test if no Issue template file exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_PullRequestTemplate(self, pat_args, snapshot):
        """
        Test if a Pull Request template file (e.g. PULL_REQUEST_TEMPLATE.md)
        exists in the repository.
        """
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == 0, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    @patch.object(
        CommunityStandardsQuery,
        "GetData",
        GetDataFunctional(".github/pull_request_template.md"),
    )
    def test_NoPullRequestTemplate(self, pat_args, snapshot):
        """Test if no Pull Request template file exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
                "--CommunityStandards-SecurityPolicy-unrequired",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_SecurityPolicy(self, pat_args, snapshot):
        """Test if a Security policy file (e.g. SECURITY.md) exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
            ],
        )

        assert result.exit_code == 0, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    @patch.object(
        CommunityStandardsQuery,
        "GetData",
        GetDataFunctional("SECURITY.md"),
    )
    def test_NoSecurityPolicy(self, pat_args, snapshot):
        """Test if no Security policy file exists in the repository."""
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--CommunityStandards-CodeOfConduct-unrequired",
                "--CommunityStandards-CodeOwners-unrequired",
                "--CommunityStandards-Contributing-unrequired",
                "--CommunityStandards-IssueTemplates-unrequired",
                "--CommunityStandards-LicenseFile-unrequired",
                "--CommunityStandards-PullRequestTemplate-unrequired",
                "--CommunityStandards-ReadMe-unrequired",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
@pytest.fixture
def args() -> list[str]:
    return [
        "--include",
        "CommunityStandards",
        "--CommunityStandards-url",
        GetGithubUrl(),
    ]


# ----------------------------------------------------------------------
@pytest.fixture
def pat_args(args) -> list[str]:
    github_pat_filename = (Path(__file__).parent / "github_pat.txt").resolve()
    CheckPATFileExists(github_pat_filename)

    return args + ["--CommunityStandards-pat", str(github_pat_filename)]



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RebaseMergeCommit object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RebaseMergeCommit(StandardEnableRequirementImpl):
    """Allow rebase merging requirement."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RebaseMergeCommit",
            False,
            "enabled",
            "settings",
            "Pull Requests",
            "Allow rebase merging",
            lambda data: data["standard"].get("allow_rebase_merge", None),
            textwrap.dedent(
                """\
                The default behavior is to not allow rebase merging.

                Reasons for this Default
                ------------------------
                - Rebase merging is not compatible with signed commits, as GitHub creates a new commit when rebasing.

                Reasons to Override this Default
                --------------------------------
                - Your repository does not require signatures.
                - You want GitHub to rebase for you as part of the pull request process when changes by others are frequent or the
                  pull request process can last for an extended period of time.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for DefaultBranchQuery.py"""

from RepoAuditor.Plugins.GitHub.DefaultBranchQuery import DefaultBranchQuery


def get_mock_request(
    status_code: int = 200,
):
    """A function which creates and returns a `mock_request` function."""

    class MockedResponse:
        def __init__(self, data, status_code):
            self._data = data
            self.status_code = status_code

        def json(self):
            return self._data

        @staticmethod
        def raise_for_status():
            pass

    def mock_request(method, url, *args, **kwargs):
        """The function to use for monkeypatching"""
        # remove leading /
        if len(url) > 0 and url[0] == "/":
            url = url[1:]

        if method.lower() == "get":
            if url == "":
                return MockedResponse({"default_branch": "main"}, status_code)
            elif url == "branches/main":
                return MockedResponse("valid-test-data", status_code)

    return mock_request


class TestDefaultBranchQuery:
    def test_GetData(self, module_data, monkeypatch):
        """Test successful GetData"""
        monkeypatch.setattr(module_data["session"], "request", get_mock_request())
        query = DefaultBranchQuery()
        query_data = query.GetData(module_data)

        assert query_data["default_branch_data"] == "valid-test-data"



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the License object."""

import textwrap
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.Impl.ValueRequirementImpl import ValueRequirementImpl


# ----------------------------------------------------------------------
class License(ValueRequirementImpl):
    """Default License as MIT Requirement."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            name="License",
            default_value="MIT License",
            github_value="License",
            get_configuration_value_func=_GetValue,
            resolution=textwrap.dedent(
                """\
                1) Create a LICENSE/LICENSE.md file in your main branch at the root of your repository.
                2) Add the appropriate licensing information in the above file.

                For detailed information about open source licenses, you can visit: https://choosealicense.com/
                """
            ),
            rationale=textwrap.dedent(
                """\
                The default behavior is to use the MIT License.

                Reasons for this Default
                ------------------------
                - The MIT License is a permissive license that allows for the use of the code in any way and
                  a reasonable default for open source software.

                Reasons to Override this Default
                --------------------------------
                - There are many good open source licenses and the MIT License may not be appropriate in all scenarios.
                """,
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(data: dict[str, Any]) -> Optional[str]:
    # Differentiate between "license is not set" and "license cannot be returned due to PAT"
    data = data["standard"]

    if "license" not in data:
        return None

    data = data["license"]
    if data is None:
        return ""

    return data["name"]



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for the EnableRulesetRequirementImpl requirement implementation."""

from RepoAuditor.Plugins.GitHub.Impl.EnableRulesetRequirementImpl import EnableRulesetRequirementImpl
from RepoAuditor.Requirement import EvaluateResult


class TestEnableRulesetRequirementImpl:
    """Unit tests for the EnableRulesetRequirementImpl class."""

    def test_Constructor(self):
        """Test the EnableRulesetRequirementImpl constructor."""
        requirement = EnableRulesetRequirementImpl(
            name="Test Name",
            enabled_by_default=False,
            dynamic_arg_name="enabled",
            github_ruleset_type="Test Ruleset",
            github_ruleset_value="test_ruleset",
            get_configuration_value_func=lambda _: True,
            resolution="Test Resolution",
            rationale="Test rationale",
        )

        assert requirement.name == "Test Name"
        assert requirement.enabled_by_default is False
        assert requirement.dynamic_arg_name == "enabled"
        assert requirement.github_ruleset_type == "Test Ruleset"

    def test_Evaluate_Disabled(self):
        """Test the _EvaluateImpl method when dynamic arg is set to `enabled: False`."""
        requirement = EnableRulesetRequirementImpl(
            name="Test Name",
            enabled_by_default=False,
            dynamic_arg_name="enabled",
            github_ruleset_type="Test Ruleset",
            github_ruleset_value="test_ruleset",
            get_configuration_value_func=lambda _: True,
            resolution="Test Resolution",
            rationale="Test rationale",
        )
        query_data = {}
        requirement_args = {"enabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        # No rule and should be disabled, hence result is success
        assert result.result == EvaluateResult.Success

    def test_Evaluate_Disabled_Alternative(self):
        """Test the _EvaluateImpl method when dynamic arg is
        set to `disabled: True` but rule is active.
        """
        requirement = EnableRulesetRequirementImpl(
            name="True By Default",
            enabled_by_default=True,
            dynamic_arg_name="disabled",
            github_ruleset_type="Test Ruleset",
            github_ruleset_value="test_ruleset",
            get_configuration_value_func=lambda rule: rule.get("type", "") == "valid_rule",
            resolution="Test Resolution",
            rationale="Test rationale",
        )
        query_data = {
            "rules": [
                {
                    "type": "valid_rule",
                    "ruleset": {
                        "name": "Valid",
                        "id": "01",
                    },
                },
            ]
        }
        requirement_args = {"disabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        # No rule and should be disabled, hence result is success
        assert result.result == EvaluateResult.Error
        assert "It should be set to False" in result.context

    def test_Evaluate_InvalidRule(self):
        """Test the _EvaluateImpl method when dynamic arg is set to `enabled: True` and the rule found is invalid."""
        requirement = EnableRulesetRequirementImpl(
            name="Test Name",
            enabled_by_default=False,
            dynamic_arg_name="enabled",
            github_ruleset_type="Test Ruleset",
            github_ruleset_value="test_ruleset",
            get_configuration_value_func=lambda rule: rule.get("type", "") == "valid rule",
            resolution="Test Resolution",
            rationale="Test rationale",
        )
        query_data = {
            "rules": [
                {
                    "type": "invalid_rule",
                    "ruleset": {
                        "name": "Invalid",
                        "id": "00",
                    },
                },
            ]
        }
        requirement_args = {"enabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error

    def test_Evaluate_ValidRule(self):
        """Test the _EvaluateImpl method when dynamic arg is set to `enabled: True` and a valid rule is found."""
        requirement = EnableRulesetRequirementImpl(
            name="Test Name",
            enabled_by_default=False,
            dynamic_arg_name="enabled",
            github_ruleset_type="Test Ruleset",
            github_ruleset_value="test_ruleset",
            get_configuration_value_func=lambda rule: rule.get("type", "") == "valid_rule",
            resolution="Test Resolution",
            rationale="Test rationale",
        )
        query_data = {
            "rules": [
                {
                    "type": "valid_rule",
                    "ruleset": {
                        "name": "Valid",
                        "id": "00",
                    },
                }
            ]
        }
        requirement_args = {"enabled": True}
        result = requirement.Evaluate(query_data, requirement_args)

        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains functionality to execute multiple Modules."""

import itertools
import sys
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Optional, cast

from dbrownell_Common import ExecuteTasks  # type: ignore[import-untyped]
from dbrownell_Common.InflectEx import inflect  # type: ignore[import-untyped]
from dbrownell_Common.Streams.Capabilities import Capabilities  # type: ignore[import-untyped]
from dbrownell_Common.Streams.DoneManager import DoneManager  # type: ignore[import-untyped]
from rich.progress import Progress, TimeElapsedColumn

from RepoAuditor.Module import EvaluateResult, ExecutionStyle, Module, OnStatusFunc
from RepoAuditor.Requirement import ReturnCode


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
@dataclass(frozen=True)
class ModuleInfo:
    """Information about a Module."""

    module: Module
    dynamic_args: dict[str, Any]
    requirement_args: dict[str, Any]


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
def Evaluate(
    module_info: ModuleInfo,
    on_status_func: OnStatusFunc,
    max_num_threads: Optional[int],
) -> list[list[Module.EvaluateInfo]]:
    """Evaluate the module given the dynamic arguments.

    Returns a list of information pertaining to module evaluation,
    with it subgrouped as lists themselves, hence a list of lists.
    """
    module_data = module_info.module.GenerateInitialData(module_info.dynamic_args)
    if module_data is None:
        return []

    return module_info.module.Evaluate(
        module_data,
        module_info.requirement_args,
        on_status_func,
        max_num_threads=max_num_threads,
    )


# ----------------------------------------------------------------------
def CalcResultInfo(
    all_eval_infos: list[list[Module.EvaluateInfo]],
    warnings_as_errors_module_names: set[str],
    ignore_warnings_module_names: set[str],
) -> tuple[int, str]:
    """Calculate the result of evaluating the modules as given in `all_eval_infos`.

    Returns a code value indicating success, warning, or error,
    as well as a string description of the result.
    If the result is a warning or error, return immediately.
    "Does Not Apply" is considered as a success.
    """
    return_code = ReturnCode.SUCCESS
    return_msg = ""

    if not all_eval_infos:
        return ReturnCode.DOESNOTAPPLY, "module does not apply"

    for eval_infos in all_eval_infos:
        for eval_info in eval_infos:
            result = eval_info.result

            if result == EvaluateResult.Warning:
                if eval_info.module.name in warnings_as_errors_module_names:
                    # Treat warnings as errors enabled so recast the result type
                    result = EvaluateResult.Error
                elif eval_info.module.name in ignore_warnings_module_names:
                    continue

            if result == EvaluateResult.Error:
                return ReturnCode.ERROR, "errors were encountered"

            if result == EvaluateResult.Warning:
                return_code, return_msg = ReturnCode.WARNING, "warnings were encountered"

    return return_code, return_msg


def Execute(
    dm: DoneManager,
    module_infos: list[ModuleInfo],
    warnings_as_errors_module_names: Optional[set[str]] = None,
    ignore_warnings_module_names: Optional[set[str]] = None,
    *,
    single_threaded: bool = False,
) -> list[list[Module.EvaluateInfo]]:
    """Execute the modules in parallel and/or sequentially."""
    if not module_infos:
        dm.WriteWarning("There are no modules to process.\n")
        return []

    warnings_as_errors_module_names = warnings_as_errors_module_names or set()
    ignore_warnings_module_names = ignore_warnings_module_names or set()
    max_num_threads = 1 if single_threaded else None

    with dm.Nested("Processing {}...".format(inflect.no("module", len(module_infos)))) as modules_dm:
        # Organize the modules into those that can be run in parallel and those that must be run
        # sequentially.
        parallel: list[tuple[int, ModuleInfo]] = []
        sequential: list[tuple[int, ModuleInfo]] = []

        for index, module_info in enumerate(module_infos):
            if module_info.module.style == ExecutionStyle.Parallel:
                parallel.append((index, module_info))
            elif module_info.module.style == ExecutionStyle.Sequential:
                sequential.append((index, module_info))
            else:
                raise RuntimeError(module_info.module.style)  # pragma: no cover

        if len(parallel) == 1:
            sequential.append(parallel[0])
            parallel = []

        # Calculate the results
        # ----------------------------------------------------------------------
        all_results: list[Optional[list[list[Module.EvaluateInfo]]]] = [None] * len(module_infos)

        if parallel:
            # ----------------------------------------------------------------------
            def Prepare(
                context: Any,  # noqa: ANN401
                on_simple_status_func: Callable[[str], None],  # noqa: ARG001
            ) -> tuple[int, ExecuteTasks.TransformTasksExTypes.TransformFuncType]:
                module_info = cast(ModuleInfo, context)
                del context

                # ----------------------------------------------------------------------
                def Transform(
                    status: ExecuteTasks.Status,
                ) -> ExecuteTasks.CompleteTransformResult:
                    # ----------------------------------------------------------------------
                    def OnStatus(num_completed: int, *args, **kwargs) -> None:
                        status.OnProgress(
                            num_completed,
                            _CreateStatusString(*args, **kwargs),
                        )

                    # ----------------------------------------------------------------------

                    evaluate_results = Evaluate(module_info, OnStatus, max_num_threads=max_num_threads)

                    result_code, result_status = CalcResultInfo(
                        evaluate_results,
                        warnings_as_errors_module_names=warnings_as_errors_module_names,
                        ignore_warnings_module_names=ignore_warnings_module_names,
                    )

                    return ExecuteTasks.CompleteTransformResult(evaluate_results, result_code, result_status)

                # ----------------------------------------------------------------------

                return module_info.module.GetNumRequirements(), Transform

            # ----------------------------------------------------------------------

            for (all_results_index, _), transformed_results in zip(
                parallel,
                ExecuteTasks.TransformTasksEx(
                    modules_dm,
                    "Processing parallel modules...",
                    [
                        ExecuteTasks.TaskData(module_info.module.name, module_info)
                        for _, module_info in parallel
                    ],
                    Prepare,
                    max_num_threads=max_num_threads,
                ),
                strict=True,
            ):
                assert all_results[all_results_index] is None
                assert isinstance(transformed_results, list), transformed_results

                all_results[all_results_index] = transformed_results

        for index, (all_results_index, module_info) in enumerate(sequential):
            with (
                modules_dm.Nested(
                    f"Processing '{module_info.module.name}' ({index + 1 + len(parallel)} of {len(module_infos)})...",
                ) as this_module_dm,
                # rich.progress needs to output to sys.stdout
                this_module_dm.YieldStdout() as stdout_context,
            ):
                stdout_context.persist_content = False

                # Technically speaking, it would be more correct to use `stdout_context.stream` here
                # rather than referencing `sys.stdout` directly, but it is really hard to work with mocked
                # stream as mocks will create mocks for everything called on the mock. Use sys.stdout
                # directly to avoid that particular problem.
                from unittest.mock import MagicMock, Mock

                assert stdout_context.stream is sys.stdout or isinstance(
                    stdout_context.stream, (Mock, MagicMock)
                ), stdout_context.stream

                with Progress(
                    *Progress.get_default_columns(),
                    TimeElapsedColumn(),
                    "{task.fields[status]}",
                    console=Capabilities.Get(sys.stdout).CreateRichConsole(sys.stdout),
                    transient=True,
                    refresh_per_second=10,
                ) as progress_bar:
                    progress_bar_task_id = progress_bar.add_task(
                        stdout_context.line_prefix,
                        status="",
                        total=module_info.module.GetNumRequirements(),
                        visible=True,
                    )

                    # ----------------------------------------------------------------------
                    def OnStatus(
                        num_completed: int,
                        num_success: int,
                        num_error: int,
                        num_warning: int,
                        num_does_not_apply: int,
                    ) -> None:
                        progress_bar.update(
                            progress_bar_task_id,  # noqa: B023
                            completed=num_completed,
                            status=_CreateStatusString(
                                num_success,
                                num_error,
                                num_warning,
                                num_does_not_apply,
                            ),
                        )

                    # ----------------------------------------------------------------------

                    evaluate_results = Evaluate(module_info, OnStatus, max_num_threads=max_num_threads)

                    assert all_results[all_results_index] is None
                    all_results[all_results_index] = evaluate_results

                    this_module_dm.result = CalcResultInfo(
                        evaluate_results,
                        warnings_as_errors_module_names=warnings_as_errors_module_names,
                        ignore_warnings_module_names=ignore_warnings_module_names,
                    )[0]

    final_results: list[list[Module.EvaluateInfo]] = []

    for results in all_results:
        if results is None:
            continue  # pragma: no cover

        final_results.append(list(itertools.chain(*results)))

    return final_results


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _CreateStatusString(
    num_success: int,
    num_error: int,
    num_warning: int,
    num_does_not_apply: int,
) -> str:
    return f"✅: {num_success} ❌: {num_error} ⚠️: {num_warning} 🚫: {num_does_not_apply}"



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the SuggestUpdatingPullRequestBranches object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class SuggestUpdatingPullRequestBranches(StandardEnableRequirementImpl):
    """Always suggest updating pull request branches."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "SuggestUpdatingPullRequestBranches",
            False,
            "enabled",
            "settings",
            "Pull Requests",
            "Always suggest updating pull request branches",
            lambda data: data["standard"].get("allow_update_branch", None),
            textwrap.dedent(
                """\
                The default behavior is to not suggest updating branches associated with pull requests within the pull request.

                Reasons for this Default
                ------------------------
                - Pull requests updated by GitHub are not compatible with signed commits, as GitHub creates a new commit when rebasing.
                - Rebasing may introduce changes that are incompatible with the current pull request.

                Reasons to Override this Default
                --------------------------------
                - Your repository does not require signatures.
                - Merge problems more insidious than conflicts are infrequent.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains types used when creating Requirements."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, IntEnum, auto
from typing import Any, Optional

from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import extension  # type: ignore[import-untyped]

from RepoAuditor.Impl.ParallelSequentialProcessor import ExecutionStyle


# ----------------------------------------------------------------------
class EvaluateResult(Enum):
    """Result of evaluating a Requirement against a set of data."""

    DoesNotApply = auto()
    Success = auto()
    Warning = auto()
    Error = auto()


class ReturnCode(IntEnum):
    """Various return codes for operations, indicating success, failure, warning, etc."""

    SUCCESS = 0
    ERROR = -1
    WARNING = 1
    DOESNOTAPPLY = 2


# ----------------------------------------------------------------------
class Requirement(ABC):
    """A single requirement that can be evaluated against a set of data."""

    # ----------------------------------------------------------------------
    # |
    # |  Public Types
    # |
    # ----------------------------------------------------------------------
    @dataclass(frozen=True)
    class EvaluateInfo:
        """Information associated with evaluating a Requirement against a set of data."""

        result: EvaluateResult
        context: Optional[str]

        resolution: Optional[str]
        rationale: Optional[str]

        requirement: "Requirement"

    # ----------------------------------------------------------------------
    @dataclass(frozen=True)
    class EvaluateImplResult:
        """Result of evaluating a Requirement."""

        result: EvaluateResult
        context: Optional[str]
        provide_resolution: bool = field(kw_only=True, default=False)
        provide_rationale: bool = field(kw_only=True, default=False)

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        description: str,
        style: ExecutionStyle,
        resolution_template: str,
        rationale_template: str,
        *,
        requires_explicit_include: bool = False,
    ) -> None:
        """Initialize the requirement with the given name, description, and style.

        If `requires_explicit_include` is True, the requirement must be explicitly included on the command line.
        """
        self.name = name
        self.description = description
        self.style = style

        self.resolution_template = resolution_template
        self.rationale_template = rationale_template

        self.requires_explicit_include = requires_explicit_include

    # ----------------------------------------------------------------------
    @extension
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Return information about dynamic arguments that the requirement can consume (often from the command line)."""

        # No dynamic arguments by default
        return {}

    # ----------------------------------------------------------------------
    def Evaluate(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> "Requirement.EvaluateInfo":
        """Evaluate the requirements given the query data and specific arguments."""
        result_info = self._EvaluateImpl(query_data, requirement_args)

        if result_info.result == EvaluateResult.Error:
            return Requirement.EvaluateInfo(
                result_info.result,
                result_info.context,
                (self.resolution_template.format(**query_data) if result_info.provide_resolution else None),
                (self.rationale_template.format(**query_data) if result_info.provide_rationale else None),
                self,
            )

        return Requirement.EvaluateInfo(result_info.result, result_info.context, None, None, self)

    # ----------------------------------------------------------------------
    # |
    # |  Private Methods
    # |
    # ----------------------------------------------------------------------
    @abstractmethod
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> "Requirement.EvaluateImplResult":
        """Perform the actual evaluation."""



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the CommandLineProcessor object."""

from dataclasses import dataclass, field
from typing import Any, Protocol

from dbrownell_Common.Streams.DoneManager import DoneManager  # type: ignore[import-untyped]

from RepoAuditor.ExecuteModules import Execute, Module, ModuleInfo


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class CommandLineProcessor:
    """Class for processing command line arguments and invoking the appropriate modules.

    Object that makes it easy to process command line arguments and invoke ExecuteModules.Execute
    in a testable way.
    """

    # ----------------------------------------------------------------------
    # |
    # |  Public Types
    # |
    # ----------------------------------------------------------------------
    class GetDynamicArgsFunc(Protocol):
        def __call__(  # noqa: D102
            self,
            dynamic_arg_definitions: dict[
                str, Any
            ],  # actual type is dict[str, TyperEx.TypeDefinitionItemType]
        ) -> dict[str, Any]: ...

    @staticmethod
    def _GetModuleMap(
        modules: list[Module],
        argument_separator: str,
    ) -> dict[str, Module]:
        """Convert modules into a lookup map.

        Returns a dict of module names mapped to the corresponding module.
        """
        module_map: dict[str, Module] = {}

        for module in modules:
            if argument_separator in module.name:
                msg = f"The module name '{module.name}' contains '{argument_separator}', which should be used as an argument separator."
                raise ValueError(msg)

            prev_added_module = module_map.get(module.name, None)
            if prev_added_module is not None:
                msg = f"The module '{module.name}' has already been defined."
                raise ValueError(msg)

            module_map[module.name] = module

        return module_map

    @staticmethod
    def _ProcessIncludes(
        module_map: dict[str, Module],
        includes: list[str],
        argument_separator: str,
    ) -> tuple[set[str], dict[str, set[str]]]:
        """Process the modules and requirements to include.

        Returns both the set of names of the included modules as well as
        the dictionary of module names mapped to the set of included requirement names.
        """
        included_modules: set[str] = set()
        included_requirements: dict[str, set[str]] = {}

        for include in includes:
            parts = include.split(argument_separator)

            this_module = module_map.get(parts[0], None)
            if this_module is None:
                msg = f"'{parts[0]}' is not a recognized module name."
                raise ValueError(msg)

            if len(parts) == 1:
                included_modules.add(parts[0])
            else:
                included_requirements.setdefault(this_module.name, set()).add(
                    argument_separator.join(parts[1:])
                )

        return included_modules, included_requirements

    @staticmethod
    def _ProcessExcludes(
        module_map: dict[str, Module],
        excludes: list[str],
        argument_separator: str,
    ) -> dict[str, set[str]]:
        """Process the modules and requirements to exclude.

        Returns both the set of names of the excluded modules as well as
        the dictionary of module names mapped to the set of excluded requirement names.
        """
        excluded_requirements: dict[str, set[str]] = {}

        for exclude in excludes:
            parts = exclude.split(argument_separator)

            this_module = module_map.get(parts[0], None)
            if this_module is None:
                msg = f"'{parts[0]}' is not a recognized module name."
                raise ValueError(msg)

            if len(parts) == 1:
                module_map.pop(parts[0])
            else:
                excluded_requirements.setdefault(this_module.name, set()).add(
                    argument_separator.join(parts[1:])
                )

        return excluded_requirements

    @staticmethod
    def _ProcessDynamicArguments(
        module_map: dict[str, Module],
        argument_separator: str,
        get_dynamic_args_func: "CommandLineProcessor.GetDynamicArgsFunc",
    ) -> dict[str, dict[str, Any]]:
        """Process the passed in dynamic arguments.

        Returns a dictionary of module names to a corresponding dictionary
        of argument names to values.
        """
        # First we create the type definitions
        dynamic_arg_definitions: dict[str, Any] = {}

        for module in module_map.values():
            # Module-level args
            for key, value in module.GetDynamicArgDefinitions().items():
                dynamic_arg_definitions[f"{module.name}{argument_separator}{key}"] = value

            # Requirement-level args
            for query in module.queries:
                for requirement in query.requirements:
                    for key, value in requirement.GetDynamicArgDefinitions().items():
                        dynamic_arg_definitions[
                            f"{module.name}{argument_separator}{requirement.name}{argument_separator}{key}"
                        ] = value

        # Now we process the definitions to get the dynamic arguments
        dynamic_args: dict[str, dict[str, Any]] = {}

        num_acceptable_dynamic_args_parts = 2

        for key, value in get_dynamic_args_func(dynamic_arg_definitions).items():
            parts = key.split(argument_separator)
            assert len(parts) >= num_acceptable_dynamic_args_parts

            if parts[0] not in module_map:
                msg = f"'{parts[0]}' is not a recognized module name."
                raise ValueError(msg)

            if len(parts) == num_acceptable_dynamic_args_parts:
                dynamic_args.setdefault(parts[0], {})[parts[1]] = value
            else:
                dynamic_args.setdefault(parts[0], {}).setdefault(None, {}).setdefault(  # type: ignore[call-overload]
                    parts[1],
                    {},
                )[argument_separator.join(parts[2:])] = value

        return dynamic_args

    # ----------------------------------------------------------------------
    # |
    # |  Public Data
    # |
    # ----------------------------------------------------------------------
    module_infos: list[ModuleInfo]
    warnings_as_error_module_names: set[str]
    ignore_warnings_module_names: set[str]
    single_threaded: bool = field(kw_only=True)

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    @classmethod
    def Create(
        cls,
        get_dynamic_args_func: "CommandLineProcessor.GetDynamicArgsFunc",
        modules: list[Module],
        includes: list[str],
        excludes: list[str],
        warnings_as_error_module_names: set[str],
        ignore_warnings_module_names: set[str],
        *,
        all_warnings_as_error: bool = False,
        ignore_all_warnings: bool = False,
        single_threaded: bool = False,
        argument_separator: str = "-",
    ) -> "CommandLineProcessor":
        """Factor method to construct a CommandLineProcessor object."""
        # Convert modules into a lookup map
        module_map: dict[str, Module] = cls._GetModuleMap(modules, argument_separator)

        del modules

        # Process includes
        included_modules, included_requirements = cls._ProcessIncludes(
            module_map, includes, argument_separator
        )

        del includes

        for module_name, module in list(module_map.items()):
            if module.requires_explicit_include and module_name not in included_modules:
                module_map.pop(module_name)
                continue

        del included_modules

        # Process excludes
        excluded_requirements: dict[str, set[str]] = cls._ProcessExcludes(
            module_map, excludes, argument_separator
        )

        del excludes

        # Add / Remove the requirements for each module
        for module_name, module in list(module_map.items()):
            module.ProcessRequirements(
                included_requirements.get(module_name, set()),
                excluded_requirements.get(module_name, set()),
            )

            if module.GetNumRequirements() == 0:
                module_map.pop(module_name)

        del included_requirements
        del excluded_requirements

        if all_warnings_as_error:
            warnings_as_error_module_names = {module.name for module in module_map.values()}
        if ignore_all_warnings:
            ignore_warnings_module_names = {module.name for module in module_map.values()}

        # Process the dynamic info
        dynamic_args: dict[str, dict[str, Any]] = cls._ProcessDynamicArguments(
            module_map, argument_separator, get_dynamic_args_func
        )

        module_infos: list[ModuleInfo] = []

        for module_name, module in module_map.items():
            module_args = dynamic_args.get(module_name, {})
            requirement_args = module_args.pop(None, None)  # type: ignore[call-overload]

            module_infos.append(ModuleInfo(module, module_args, requirement_args))

        return cls(
            module_infos,
            warnings_as_error_module_names,
            ignore_warnings_module_names,
            single_threaded=single_threaded,
        )

    # ----------------------------------------------------------------------
    def __call__(  # noqa: D102
        self,
        dm: DoneManager,
    ) -> list[list[Module.EvaluateInfo]]:
        return Execute(dm, self.module_infos)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the EnableRequirementImpl object."""

from collections.abc import Callable
from typing import Any, Optional

from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.Impl.EnableRequirementImpl import EnableRequirementImpl
from RepoAuditor.Requirement import EvaluateResult, Requirement


# ----------------------------------------------------------------------
class EnableRulesetRequirementImpl(EnableRequirementImpl):
    """Object that implements ruleset settings that can be enabled/disabled."""

    # ----------------------------------------------------------------------
    def __init__(  # noqa: PLR0913
        self,
        name: str,
        enabled_by_default: bool,  # noqa: FBT001
        dynamic_arg_name: str,
        github_ruleset_type: str,
        github_ruleset_value: str,
        get_configuration_value_func: Callable[[dict[str, Any]], Optional[bool]],
        resolution: str,
        rationale: str,
        subject: Optional[str] = None,
        *,
        requires_explicit_include: bool = False,
        unset_set_terminology: tuple[str, str] = ("disabled", "enabled"),
        missing_value_is_warning: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            enabled_by_default=enabled_by_default,
            dynamic_arg_name=dynamic_arg_name,
            github_settings_value=github_ruleset_value,
            get_configuration_value_func=get_configuration_value_func,
            resolution=resolution,
            rationale=rationale,
            subject=subject,
            requires_explicit_include=requires_explicit_include,
            unset_set_terminology=unset_set_terminology,
            missing_value_is_warning=missing_value_is_warning,
        )

        self.github_ruleset_type = github_ruleset_type

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        expected_rule_enabled = self.enabled_by_default

        if requirement_args[self.dynamic_arg_name]:
            expected_rule_enabled = not expected_rule_enabled
            provide_rationale = False
        else:
            provide_rationale = True

        query_data["__expected_value"] = "Enable" if expected_rule_enabled else "Disable"

        # Get the rules for the specified branch
        rules = query_data.get("rules", [])

        # Check if the rule is checked in the ruleset
        rule_enabled = False
        ruleset = None
        for rule in rules:
            # If the rule is present, it is checked
            if self.get_configuration_value_func(rule):
                rule_enabled = True
                # Get the ruleset associated with the rule
                ruleset = rule["ruleset"]
                break

        # If rule is active and we want it enabled,
        # or rule is inactive and we want it disabled,
        # return success.
        if rule_enabled == expected_rule_enabled:
            context = (
                f"{self.github_settings_value} is enabled in Ruleset '{ruleset['name']}'"
                if expected_rule_enabled
                else f"{self.github_settings_value} is disabled"
            )
            return self.EvaluateImplResult(
                EvaluateResult.Success,
                context,
            )

        # When rule is active but we want it disabled
        if rule_enabled:
            assert expected_rule_enabled is False

            return self.EvaluateImplResult(
                EvaluateResult.Error,
                f"{self.github_settings_value} rule active in ruleset '{ruleset['name']}'. It should be set to {expected_rule_enabled}",
                provide_resolution=True,
                provide_rationale=provide_rationale,
            )

        # Rule is inactive but we want it on
        assert rule_enabled is False
        assert expected_rule_enabled is True

        return self.EvaluateImplResult(
            EvaluateResult.Error,
            f"No active branch ruleset with {self.github_settings_value} found",
            provide_resolution=True,
            provide_rationale=provide_rationale,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for GitHub/Impl/ValueRequirementImpl.py"""

import pytest

from RepoAuditor.Plugins.GitHub.Impl.ValueRequirementImpl import DoesNotApplyResult, ValueRequirementImpl
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def requirement():
    return ValueRequirementImpl(
        "TestValueRequirement",
        "42",
        "Test Value Requirement",
        lambda data: data.get("result", None),
        "",
        "",
    )


@pytest.fixture
def requirement_does_not_apply_result():
    return ValueRequirementImpl(
        "TestValueRequirement",
        "42",
        None,
        lambda data: DoesNotApplyResult(""),
        "",
        "",
    )


class TestValueRequirementImpl:
    """Test the ValueRequirementImpl class."""

    def test_ResultIsNone(self, requirement):
        """Test pathway where result value is None."""
        query_data = {"result": None}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data" in result.context

    def test_ResultDoesntMatchExpectedValue(self, requirement):
        """Test pathway where result value and expected value don't match."""
        query_data = {"result": "43"}
        requirement_args = {"value": "40"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "'Test Value Requirement' must be set to '40' (it is currently set to '43')." in result.context

    def test_Successful(self, requirement):
        """Test successful."""
        query_data = {"result": "42"}
        requirement_args = {"value": "42"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

    def test_SuccessfulGetDynamicArgDefinitions(self, requirement):
        """Test successful with GetDynamicArgDefinitions/"""
        query_data = {"result": "42"}
        requirement_args = {}
        for key, value in requirement.GetDynamicArgDefinitions().items():
            requirement_args[key] = value[1].default
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

    def test_MissingValueIsWarning(self):
        """Test for result when missing_value_is_warning is False."""
        requirement = ValueRequirementImpl(
            "TestValueRequirement",
            "42",
            "Test Value Requirement",
            lambda data: data.get("result", None),
            "",
            "",
            missing_value_is_warning=False,
        )

        query_data = {"result": None}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_ExpectedValueIsDefault(self, requirement_does_not_apply_result):
        """Test pathway where expected value is default."""
        query_data = {"result": "43"}
        requirement_args = {"value": "42"}
        result = requirement_does_not_apply_result.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply

    def test_ResultDoesNotApply(self, requirement_does_not_apply_result):
        """Test pathway where result is DoesNotApplyResult."""
        query_data = {"result": "42"}
        requirement_args = {"value": "43"}
        result = requirement_does_not_apply_result.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "the entity cannot be set to '43' because " in result.context



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the PullRequestTemplate requirement."""

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class PullRequestTemplate(ExistsRequirementImpl):
    """Validates that a pull request template is configured."""

    def __init__(self) -> None:
        super().__init__(
            "PullRequestTemplate",
            "PULL_REQUESTS_TEMPLATE",
            [
                ".github/PULL_REQUEST_TEMPLATE.md",
                "docs/PULL_REQUEST_TEMPLATE.md",
                "PULL_REQUEST_TEMPLATE.md",
                ".github/pull_request_template.md",
                "docs/pull_request_template.md",
                "pull_request_template.md",
                # Directory with multiple templates
                ".github/PULL_REQUEST_TEMPLATE",
            ],
            "Create a pull request template file in one of these locations: .github/PULL_REQUEST_TEMPLATE.md, .github/pull_request_template.md, docs/PULL_REQUEST_TEMPLATE.md, or PULL_REQUEST_TEMPLATE.md",
            "Pull request templates help standardize code contributions and expedite the review process by ensuring consistent PR format.",
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the DefaultBranchQuery object."""

from typing import Any, Optional

from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.DefaultBranchRequirements.Protected import Protected
from RepoAuditor.Query import ExecutionStyle, Query


# ----------------------------------------------------------------------
class DefaultBranchQuery(Query):
    """Query with requirements that operate on branch information."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "DefaultBranchQuery",
            ExecutionStyle.Parallel,
            [
                Protected(),
            ],
        )

    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Get the data from an API session."""
        # Get the default branch name
        response = module_data["session"].get("")

        response.raise_for_status()
        response = response.json()

        module_data["default_branch"] = response["default_branch"]

        # Get the information for the default branch
        response = module_data["session"].get(f"branches/{module_data['default_branch']}")

        response.raise_for_status()
        response = response.json()

        module_data["default_branch_data"] = response

        return module_data



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the SupportIssues object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class SupportIssues(StandardEnableRequirementImpl):
    """Support for Issues."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "SupportIssues",
            True,
            "disabled",
            "settings",
            "Features",
            "Issues",
            lambda data: data["standard"].get("has_issues", None),
            rationale=textwrap.dedent(
                """\
                No rationale for this default.
                """,
            ),
            subject="Support for Issues",
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for RequireApprovals.py"""

import pytest

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireApprovals import RequireApprovals
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "branch": "main",
        "branch_protection_data": {
            "required_pull_request_reviews": {
                "required_approving_review_count": "1",
            },
        },
    }


@pytest.fixture
def requirement():
    return RequireApprovals()


class TestRequireApprovals:
    def test_required_pull_request_reviews_missing(self, requirement, query_data):
        """Test when `required_pull_request_reviews` is missing"""
        query_data["branch_protection_data"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_required_pull_request_reviews_none(self, requirement, query_data):
        """Test when `required_pull_request_reviews` is None"""
        query_data["branch_protection_data"]["required_pull_request_reviews"] = None
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_required_approving_review_count_missing(self, requirement, query_data):
        """Test when `required_approving_review_count` is missing"""
        query_data["branch_protection_data"]["required_pull_request_reviews"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_required_approving_review_count_none(self, requirement, query_data):
        """Test when `required_approving_review_count` is None"""
        query_data["branch_protection_data"]["required_pull_request_reviews"][
            "required_approving_review_count"
        ] = None
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_IncorrectValue(self, requirement, query_data):
        """Test incorrect"""
        query_data["branch_protection_data"]["required_pull_request_reviews"][
            "required_approving_review_count"
        ] = "2"
        requirement_args = {"value": "1"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert (
            "'Require a pull request before merging -> Require approvals' must be set to '1' (it is currently set to '2')."
            in result.context
        )

    def test_Successful(self, requirement, query_data):
        """Test successful"""
        query_data["branch_protection_data"]["required_pull_request_reviews"][
            "required_approving_review_count"
        ] = "1"
        requirement_args = {"value": "1"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for DismissStalePullRequestApprovals.py"""

import pytest

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.DismissStalePullRequestApprovals import (
    DismissStalePullRequestApprovals,
)
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "branch": "main",
        "branch_protection_data": {
            "required_pull_request_reviews": {
                "dismiss_stale_reviews": True,
            },
        },
    }


@pytest.fixture
def requirement():
    return DismissStalePullRequestApprovals()


class TestDismissStalePullRequestApprovals:
    def test_required_pull_request_reviews_missing(self, requirement, query_data):
        """Test when `required_pull_request_reviews` is missing."""
        query_data["branch_protection_data"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_required_pull_request_reviews_none(self, requirement, query_data):
        """Test when `required_pull_request_reviews` is None"""
        query_data["branch_protection_data"]["required_pull_request_reviews"] = None
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_dismiss_stale_reviews_missing(self, requirement, query_data):
        """Test when `dismiss_stale_reviews` is missing"""
        query_data["branch_protection_data"]["required_pull_request_reviews"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_dismiss_stale_reviews_none(self, requirement, query_data):
        """Test when `dismiss_stale_reviews` is None"""
        query_data["branch_protection_data"]["required_pull_request_reviews"] = {
            "dismiss_stale_reviews": None
        }
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_successful(self, requirement, query_data):
        """Test successful"""
        query_data["branch_protection_data"]["required_pull_request_reviews"]["dismiss_stale_reviews"] = True
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for ClassicBranchProtectionQuery.py"""

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionQuery import ClassicBranchProtectionQuery


def get_mock_request(
    status_code: int = 200,
    protected_branch: bool = True,
    rulesets_exist: bool = True,
):
    """A function which creates and returns a `mock_request` function."""

    class MockedResponse:
        def __init__(self, data, status_code):
            self._data = data
            self.status_code = status_code

        def json(self):
            return self._data

        @staticmethod
        def raise_for_status():
            pass

    def mock_request(method, url, *args, **kwargs):
        """The function to use for monkeypatching"""
        # remove leading /
        if len(url) > 0 and url[0] == "/":
            url = url[1:]

        if method.lower() == "get":
            if url == "":
                return MockedResponse({"default_branch": "main"}, status_code)
            elif url == "branches/main":
                return MockedResponse({"protected": protected_branch}, status_code)
            elif url == "branches/main/protection":
                return MockedResponse("valid-test-data", status_code)
            elif url == "rules/branches/main":
                if rulesets_exist:
                    return MockedResponse("valid-rules", status_code=200)
                else:
                    return MockedResponse("", status_code=200)

    return mock_request


class TestClassicBranchProtectionQuery:
    def test_GetData(self, module_data, monkeypatch):
        """Test successful GetData"""
        monkeypatch.setattr(module_data["session"], "request", get_mock_request())
        query = ClassicBranchProtectionQuery()
        query_data = query.GetData(module_data)

        assert query_data["branch_protection_data"] == "valid-test-data"

    def test_NotProtected(self, module_data, monkeypatch):
        """Test where branch is not protected."""
        monkeypatch.setattr(module_data["session"], "request", get_mock_request(protected_branch=False))
        query = ClassicBranchProtectionQuery()
        query_data = query.GetData(module_data)

        assert query_data is None

    def test_RulesetProtected(self, module_data, monkeypatch):
        """Test when the branch is protected by rulesets."""
        monkeypatch.setattr(module_data["session"], "request", get_mock_request(status_code=404))
        query = ClassicBranchProtectionQuery()
        query_data = query.GetData(module_data)

        assert query_data is None

    def test_RulesetProtectedNoRules(self, module_data, monkeypatch):
        """Test when the branch is protected by rulesets, but no rules exist."""
        monkeypatch.setattr(
            module_data["session"], "request", get_mock_request(status_code=404, rulesets_exist=False)
        )
        # Set github_pat to None to test a pathway
        module_data["session"].github_pat = None
        query = ClassicBranchProtectionQuery()
        query_data = query.GetData(module_data)

        assert query_data is None



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for GitHubSession"""

from pathlib import Path

import pytest
import requests

from RepoAuditor.Plugins.GitHubBase.Module import _GitHubSession


@pytest.fixture
def github_pat(fs):
    """Fixture for GitHub PAT file, which creates a fake PAT file and returns the path."""
    pat_path = Path(__file__).parent / "dummy_github_pat.txt"
    fs.create_file(pat_path, contents="github_path_abcdefghijklmnop")
    return pat_path


def mock_request(
    self,
    method,
    url,
    *args,
    **kwargs,
):
    """Mocked request function."""
    r = requests.Response()
    r.status_code = 200
    r.method = method
    r.url = url
    return r


class TestGitHubSession:
    """Unit tests for the _GitHubSession class."""

    @classmethod
    def setup_class(cls):
        """Setup values for each test."""
        cls.github_url = "https://github.com/gt-sse-center/RepoAuditor"

    def test_Construct(self, github_pat):
        """Test constructor."""
        session = _GitHubSession(github_url=self.github_url, github_pat=github_pat)

        assert session.github_url == self.github_url

    def test_ConstructTrailingSlash(self, github_pat):
        """Test with trailing slash for URL."""
        session = _GitHubSession(github_url=self.github_url + "/", github_pat=github_pat)

        assert session.github_url == self.github_url

    def test_InvalidRepository(self, github_pat):
        """Test with invalid repository URL."""
        with pytest.raises(ValueError):
            _GitHubSession(
                github_url="https://github.com/gt-sse-center/RepoAuditor/123", github_pat=github_pat
            )

    def test_EnterpriseUrl(self, github_pat):
        """Test with enterprise URL."""
        github_enterprise_url = "https://github.gatech.edu/gt-sse-center/RepoAuditor"
        session = _GitHubSession(github_url=github_enterprise_url, github_pat=github_pat)
        assert session.github_url == github_enterprise_url

    def test_Request(self, github_pat, monkeypatch):
        """Test the _GitHubSession.request method."""
        session = _GitHubSession(github_url=self.github_url, github_pat=github_pat)

        monkeypatch.setattr(requests.Session, "request", mock_request)
        r = session.request("GET", "test")

        assert r.url == "https://api.github.com/repos/gt-sse-center/RepoAuditor/test"

    def test_RequestLeadingSlash(self, github_pat, monkeypatch):
        """Test the _GitHubSession.request method."""
        session = _GitHubSession(github_url=self.github_url, github_pat=github_pat)

        monkeypatch.setattr(requests.Session, "request", mock_request)
        r = session.request("GET", "/test")

        assert r.url == "https://api.github.com/repos/gt-sse-center/RepoAuditor/test"



import pytest

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.RequireUpToDateBranches import (
    RequireUpToDateBranches,
)
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "branch": "main",
        "branch_protection_data": {
            "required_status_checks": {
                "strict": False,
            },
        },
    }


@pytest.fixture
def requirement():
    return RequireUpToDateBranches()


class TestRequireUpToDateBranches:
    def test_required_status_checks_missing(self, requirement, query_data):
        """Test when `required_status_checks` is missing"""
        query_data["branch_protection_data"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_required_status_checks_none(self, requirement, query_data):
        """Test when `required_status_checks` is None"""
        query_data["branch_protection_data"] = {"required_status_checks": None}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_strict_missing(self, requirement, query_data):
        """Test when `strict` is missing"""
        query_data["branch_protection_data"]["required_status_checks"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_strict_none(self, requirement, query_data):
        """Test when `strict` is None"""
        query_data["branch_protection_data"]["required_status_checks"] = {"strict": None}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_IncorrectValue(self, requirement, query_data):
        """Test incorrect"""
        query_data["branch_protection_data"]["required_status_checks"]["strict"] = False
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert (
            "'Require status checks to pass before merging -> Require branches to be up to date before merging' must be set to 'True' (it is currently set to 'False')."
            in result.context
        )

    def test_Successful(self, requirement, query_data):
        """Test successful"""
        query_data["branch_protection_data"]["required_status_checks"]["strict"] = False
        requirement_args = {"disabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the GitHubModule object."""

from RepoAuditor.Module import ExecutionStyle
from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionQuery import ClassicBranchProtectionQuery
from RepoAuditor.Plugins.GitHub.DefaultBranchQuery import DefaultBranchQuery
from RepoAuditor.Plugins.GitHub.RulesetQuery import RulesetQuery
from RepoAuditor.Plugins.GitHub.StandardQuery import StandardQuery
from RepoAuditor.Plugins.GitHubBase.Module import GitHubBaseModule


# ----------------------------------------------------------------------
class GitHubModule(GitHubBaseModule):
    """Module for validating GitHub repository configuration settings."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "GitHub",
            "Validates GitHub configuration settings.",
            ExecutionStyle.Parallel,
            [
                StandardQuery(),
                DefaultBranchQuery(),
                ClassicBranchProtectionQuery(),
                RulesetQuery(),
            ],
            requires_explicit_include=True,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the LicenseFile requirement."""

import textwrap

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import (
    ExistsRequirementImpl,
)


class LicenseFile(ExistsRequirementImpl):
    """Validates that a CODE_OF_CONDUCT file is configured."""

    def __init__(self) -> None:
        super().__init__(
            "LicenseFile",
            "LICENSE",
            [
                "LICENSE",
                "LICENSE.md",
                "LICENSE.txt",
                "LICENSE.rst",
            ],
            textwrap.dedent(
                """\
                1) Create a LICENSE file in in the root directory.

                2) It can have any of the following extensions:
                   - [blank] e.g. `LICENSE`
                   - LICENSE.md
                   - LICENSE.rst
                   - LICENSE.txt
                """
            ),
            textwrap.dedent(
                """\
                A software license tells others what they can and can't do with your source code.

                https://choosealicense.com/ can help decide which license
                best meets your repository's needs.
                """
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the WebCommitSignoff object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class WebCommitSignoff(StandardEnableRequirementImpl):
    """Require contributors to sign off on web-based commits."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "WebCommitSignoff",
            True,
            "disabled",
            "settings",
            "General",
            "Require contributors to sign off on web-based commits",
            lambda data: data["standard"].get("web_commit_signoff_required", None),
            rationale=textwrap.dedent(
                """\
                The default behavior is to require contributors to sign off on web-based commits.

                Reasons for this Default
                ------------------------
                - All changes (regardless of where they were made) should go through the same validation process.

                Reasons to Override this Default
                --------------------------------
                - Changes made via the web interface are considered to be benign and should not be subject to
                  the standard validation process.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the SecurityPolicy requirement."""

import textwrap

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class SecurityPolicy(ExistsRequirementImpl):
    """Validates that a security policy is configured."""

    def __init__(self) -> None:
        super().__init__(
            "SecurityPolicy",
            "SECURITY",
            [
                ".github/SECURITY.md",
                "docs/SECURITY.md",
                "SECURITY.md",
            ],
            textwrap.dedent(
                """\
                1) Create a security policy file in one of these locations:
                   - .github/SECURITY.md
                   - docs/SECURITY.md
                   - SECURITY.md

                Example policy can be found at:
                https://github.com/gt-sse-center/RepoAuditor/blob/main/SECURITY.md
                """
            ),
            textwrap.dedent(
                """\
                Security policies help users report vulnerabilities responsibly.

                Benefits:
                - Clear security reporting process
                - Responsible disclosure guidelines
                - Better security management
                """
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the DependabotSecurityUpdates object."""

import textwrap
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class DependabotSecurityUpdates(StandardEnableRequirementImpl):
    """Requirement of Dependabot security updates."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "DependabotSecurityUpdates",
            True,
            "disabled",
            "settings/security_analysis",
            "Dependabot",
            "Dependabot security updates",
            _GetValue,
            textwrap.dedent(
                """\
                The default behavior is to ensure Dependabot security updates are enabled.

                Reasons for this Default
                ------------------------
                - Increases the overall security of the repository by automatically applying security updates.

                Reasons to Override this Default
                --------------------------------
                - Dependabot security updates are not supported for the repository or by the organization.
                - A manual test pass is required before changes can be deployed.
                """,
            ),
            unset_set_terminology=("disabled", "enabled"),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(data: dict[str, Any]) -> Optional[bool]:
    result = (
        data["standard"]
        .get("security_and_analysis", {})
        .get("dependabot_security_updates", {})
        .get("status", None)
    )

    if result is None:
        return None

    return result == "enabled"



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the EnsureStatusChecks object."""

import textwrap
from typing import Any

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
class EnsureStatusChecks(Requirement):
    """Ensure that status checks have been enabled for the branch."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "EnsureStatusChecks",
            "Ensure that status checks have been enabled for the branch.",
            ExecutionStyle.Parallel,
            textwrap.dedent(
                """\
                1) Visit '{session.github_url}/settings/branches'
                2) Locate the 'Branch protection rules' section
                3) Click the 'Edit' button next to the branch '{branch}'
                4) Locate the 'Protect matching branches' section
                5) Locate the 'Require status checks to pass before merging' section
                6) Add status checks that must pass before merging
                """,
            ),
            textwrap.dedent(
                """\
                The default behavior is to require the specification of at least one status check that must pass before merging.

                Reasons for this Default
                ------------------------
                - A Continuous Integration process must be gated by some activity, action, or work flow.

                Reasons to Override this Default
                --------------------------------
                - Code reviews are sufficient to ensure the quality of pull requests before they are merged into the branch
                  (this is not recommended).
                """,
            ),
        )

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            "disabled": (
                bool,
                typer.Option(
                    default=False,
                    help="If set, status checks will not be enforced.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        if requirement_args["disabled"]:
            return Requirement.EvaluateImplResult(
                EvaluateResult.DoesNotApply,
                "The status check requirement has been explicitly disabled.",
            )

        settings = query_data["branch_protection_data"].get("required_status_checks", None)
        if settings is None:
            return Requirement.EvaluateImplResult(EvaluateResult.DoesNotApply, None)

        settings = settings.get("checks", None)
        if not settings:
            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                "No status checks are configured.",
                provide_resolution=True,
                provide_rationale=True,
            )

        return Requirement.EvaluateImplResult(EvaluateResult.Success, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for RulesetQuery.py"""

from RepoAuditor.Plugins.GitHub.RulesetQuery import RulesetQuery


def get_mock_request(
    status_code: int = 200,
):
    """A function which creates and returns a `mock_request` function."""

    class MockedResponse:
        def __init__(self, data, status_code):
            self._data = data
            self.status_code = status_code

        def json(self):
            return self._data

        @staticmethod
        def raise_for_status():
            pass

    def mock_request(method, url, *args, **kwargs):
        """The function to use for monkeypatching"""
        # remove leading /
        if len(url) > 0 and url[0] == "/":
            url = url[1:]

        if method.lower() == "get":
            if url == "":
                return MockedResponse({"default_branch": "main"}, status_code)
            elif url == "rules/branches/main":
                return MockedResponse(
                    [
                        {
                            "ruleset_id": 117,
                        },
                    ],
                    status_code,
                )
            elif url == "rulesets/117":
                return MockedResponse("valid-test-data", status_code)

    return mock_request


class TestRulesetQuery:
    def test_GetData(self, module_data, monkeypatch):
        """Test the GetData method."""
        monkeypatch.setattr(module_data["session"], "request", get_mock_request())
        query = RulesetQuery()
        query_data = query.GetData(module_data)

        assert len(query_data["rules"]) == 1
        assert query_data["rules"][0]["ruleset_id"] == 117
        assert query_data["rules"][0]["ruleset"] == "valid-test-data"



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the Protected object."""

import textwrap
from typing import Any

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.Impl.Common import CreateIncompleteDataResult
from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
class Protected(Requirement):
    """Requirement to ensure that the mainline branch is protected."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        # Note that protected is set for the branch when creating a branch ruleset or a classic
        # branch protection rule.
        super().__init__(
            "Protected",
            "Ensures that the mainline branch is protected.",
            ExecutionStyle.Parallel,
            textwrap.dedent(
                """\
                1) Visit '{session.github_url}/settings/branches'
                2) Locate the 'Branch protection rules' section
                3) Protect the branch with a ruleset or classic branch protection rule
                """,
            ),
            textwrap.dedent(
                """\
                The default behavior is to protect the mainline branch.

                Reasons for this Default
                ------------------------
                - The mainline branch is the most important branch in a repository and should be protected.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            "disabled": (
                bool,
                typer.Option(
                    default=False,
                    help="Allow an unprotected mainline branch.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        is_protected = query_data["default_branch_data"].get("protected", None)
        if is_protected is None:
            return CreateIncompleteDataResult()

        if requirement_args["disabled"]:
            expected_value = False
            provide_rationale = False
        else:
            expected_value = True
            provide_rationale = True

        if is_protected != expected_value:
            if expected_value:
                modifier = ""
                current_modifier = " not"
            else:
                modifier = " not"
                current_modifier = ""

            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                f"The mainline branch must{modifier} be protected (it is currently{current_modifier} protected).",
                provide_resolution=True,
                provide_rationale=provide_rationale,
            )

        return Requirement.EvaluateImplResult(EvaluateResult.Success, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the DeleteHeadBranches object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class DeleteHeadBranches(StandardEnableRequirementImpl):
    """Requirement to delete head branches."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "DeleteHeadBranches",
            True,
            "disabled",
            "settings",
            "Pull Requests",
            "Automatically delete head branches",
            lambda data: data["standard"].get("delete_branch_on_merge", None),
            textwrap.dedent(
                """\
                The default behavior is to automatically delete head branches once they have been merged into the mainline branch.

                Reasons for this Default
                ------------------------
                - Long-lived branches make integration more difficult due to changes that accumulate over time.

                Reasons to Override this Default
                --------------------------------
                - You support release branches and may want to merge changes from this release branch into the mainline branch
                  (although, it is possible to workaround this issue by creating a pull request from a temporary branch that
                  includes cherry-picked changes from the release branch).
                """,
            ),
        )



# ----------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Audits a repository against a set of requirements."""

import os
import sys
import textwrap
import traceback
from typing import Annotated, Optional

import pluggy
import typer
from click.exceptions import UsageError
from dbrownell_Common import (
    TextwrapEx,  # type: ignore [import-untyped]
    TyperEx,  # type: ignore [import-untyped]
)
from dbrownell_Common.Streams.DoneManager import DoneManager
from dbrownell_Common.Streams.DoneManager import Flags as DoneManagerFlags  # type: ignore [import-untyped]
from typer.core import TyperGroup  # type: ignore [import-untyped]
from typer_config.decorators import use_yaml_config

from RepoAuditor import APP_NAME, Plugin, __version__
from RepoAuditor.CommandLineProcessor import CommandLineProcessor, Module
from RepoAuditor.Display import DisplayResults

# ----------------------------------------------------------------------
ARGUMENT_SEPARATOR = "-"


# ----------------------------------------------------------------------
class NaturalOrderGrouper(TyperGroup):
    """Group commands in natural order."""

    # ----------------------------------------------------------------------
    def list_commands(self, *args, **kwargs) -> list[str]:  # noqa: ARG002  # pragma: no cover
        """Return a list of all the commands, to be sorted in a natural ordering."""
        return list(self.commands.keys())


# ----------------------------------------------------------------------
app = typer.Typer(
    cls=NaturalOrderGrouper,
    help=__doc__,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)


# ----------------------------------------------------------------------
def _GetModules() -> list[Module]:
    plugin_manager = pluggy.PluginManager(APP_NAME)

    plugin_manager.add_hookspecs(Plugin)
    plugin_manager.load_setuptools_entrypoints(APP_NAME)

    return plugin_manager.hook.GetModule()


_all_modules = _GetModules()
del _GetModules


# ----------------------------------------------------------------------
def TypeInfoToString(
    arg_name: str,
    type_info: TyperEx.TypeDefinitionItemType,
) -> str:
    """Convert type information from command line arguments to a string representation.

    Args:
        arg_name (str): The name of the command line argument.
        type_info (TyperEx.TypeDefinitionItemType): Type information for the accepted value.

    Returns:
        str: String description of the command line argument.

    """
    if isinstance(type_info, TyperEx.TypeDefinitionItem):
        python_type = type_info.python_type
        parameter_info = type_info.parameter_info
    elif isinstance(type_info, tuple):
        python_type = type_info[0]
        parameter_info = type_info[1]
    else:
        python_type = type_info
        parameter_info = None

    return "    {arg_name:<50} {type_description:<7} {help}".format(
        arg_name=arg_name,
        type_description=python_type.__name__,
        help="" if parameter_info is None else parameter_info.help,
    )


# ----------------------------------------------------------------------
def _HelpEpilog() -> str:
    content: list[str] = []

    # ----------------------------------------------------------------------

    for module in _all_modules:
        # Get the module arguments
        module_arguments: list[str] = []

        for arg_name, type_info in module.GetDynamicArgDefinitions().items():
            module_arguments.append(
                TypeInfoToString(
                    f"--{module.name}{ARGUMENT_SEPARATOR}{arg_name}",
                    type_info,
                )
            )

        final_module_arguments = TextwrapEx.Indent(
            "\n".join(module_arguments),
            4,
            skip_first_line=True,
        )

        # Get the requirement arguments
        requirement_arguments: list[str] = []

        for query in module.queries:
            for requirement in query.requirements:
                for key, value in requirement.GetDynamicArgDefinitions().items():
                    requirement_arguments.append(
                        TypeInfoToString(
                            f"--{module.name}{ARGUMENT_SEPARATOR}{requirement.name}{ARGUMENT_SEPARATOR}{key}",
                            value,
                        ),
                    )

        final_requirement_arguments = TextwrapEx.Indent(
            "\n".join(requirement_arguments),
            4,
            skip_first_line=True,
        )

        content.append(
            textwrap.dedent(
                f"""\
                {"-" * len(module.name)}
                {module.name}
                {"-" * len(module.name)}
                {module.description}

                Command Line Arguments:
                    {final_module_arguments}

                Optional Arguments:
                    {final_requirement_arguments}
                """,
            ),
        )

    return (
        textwrap.dedent(
            """\
        Module Information
        ==================

        {}

        Command Line Examples
        =====================
        repo_auditor --include GitHub --GitHub-url https://github.com/gt-sse-center/RepoAuditor --GitHub-pat <PAT or path to file containing PAT>

        """,
        )
        .format(
            "\n".join(content),
        )
        .replace("\n", "\n\n")
    )


# ----------------------------------------------------------------------
def _VersionCallback(value: bool) -> None:  # noqa: FBT001
    if value:
        sys.stdout.write(f"RepoAuditor v{__version__}\n")
        raise typer.Exit()


# ----------------------------------------------------------------------
@app.command(
    "EntryPoint",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    help=__doc__,
    epilog=_HelpEpilog(),
    no_args_is_help=False,
)
@use_yaml_config()
def EntryPoint(  # noqa: PLR0913  # pragma: no cover
    ctx: typer.Context,
    version: Annotated[  # noqa: ARG001, FBT002
        bool,
        typer.Option(
            "--version",
            help="Display the version of this tool and exit.",
            callback=_VersionCallback,
            is_eager=True,
        ),
    ] = False,
    includes: Annotated[
        Optional[list[str]],
        typer.Option(
            "--include",
            help=f"Module or requirement names to explicitly include in the execution; like other command line arguments, requirement names must include the module name as a prefix (e.g. 'ModuleName{ARGUMENT_SEPARATOR}RequirementName'). This value can be provided multiple times.",
        ),
    ] = None,
    excludes: Annotated[
        Optional[list[str]],
        typer.Option(
            "--exclude",
            help=f"Module or requirement names to exclude from execution; like other command line arguments, requirement names must include the module name as a prefix (e.g. 'ModuleName{ARGUMENT_SEPARATOR}RequirementName'). This value can be provided multiple times.",
        ),
    ] = None,
    warnings_as_error: Annotated[
        Optional[list[str]],
        typer.Option(
            "--warnings-as-error",
            help="Name of a module whose warnings should be treated as errors. This value can be provided multiple times.",
        ),
    ] = None,
    ignore_warnings: Annotated[
        Optional[list[str]],
        typer.Option(
            "--ignore-warnings",
            help="Name of a module whose warnings should be ignored. This value can be provided multiple times.",
        ),
    ] = None,
    all_warnings_as_error: Annotated[  # noqa: FBT002
        bool, typer.Option("--all-warnings-as-error", help="Treat all warnings as errors.")
    ] = False,
    ignore_all_warnings: Annotated[  # noqa: FBT002
        bool, typer.Option("--ignore-all-warnings", help="Ignore all warnings.")
    ] = False,
    single_threaded: Annotated[  # noqa: FBT002
        bool,
        typer.Option("--single-threaded", help="Do not use multiple threads when evaluating requirements."),
    ] = False,
    no_resolution: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--no-resolution",
            help="Do not display resolution information for requirements that are not successful.",
        ),
    ] = False,
    no_rationale: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--no-rationale",
            help="Do not display rationale information for requirements that are not successful.",
        ),
    ] = False,
    verbose: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--verbose",
            help="Write verbose information to the terminal.",
        ),
    ] = False,
    debug: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--debug",
            help="Write debug information to the terminal.",
        ),
    ] = False,
) -> None:
    """Entry point for the command line interface."""
    # Initialize lists to avoid mutable default arguments
    includes = includes or []
    excludes = excludes or []
    warnings_as_error = warnings_as_error or []
    ignore_warnings = ignore_warnings or []

    with DoneManager.CreateCommandLine(
        sys.stdout,
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        try:
            executor = CommandLineProcessor.Create(
                lambda dynamic_arg_definitions: TyperEx.ProcessDynamicArgs(ctx, dynamic_arg_definitions),
                _all_modules,
                includes,
                excludes,
                set(warnings_as_error),
                set(ignore_warnings),
                all_warnings_as_error=all_warnings_as_error,
                ignore_all_warnings=ignore_all_warnings,
                single_threaded=single_threaded,
                argument_separator=ARGUMENT_SEPARATOR,
            )

        except Exception as ex:
            if dm.is_debug:
                raise

            raise UsageError(str(ex)) from ex

        try:
            all_results = executor(dm)

            dm.WriteLine("\n\n")

            # Try to set panel width
            try:
                panel_width = min(
                    os.get_terminal_size().columns,
                    dm.capabilities.DEFAULT_COLUMNS,
                )
            except OSError:
                # Exception if not connected to a terminal
                # in which case default to None
                panel_width = None

            DisplayResults(
                dm,
                all_results,
                display_resolution=not no_resolution,
                display_rationale=not no_rationale,
                panel_width=panel_width,
            )
        except Exception as ex:
            error = traceback.format_exc() if dm.is_debug else str(ex)
            dm.WriteError(error)


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
if __name__ == "__main__":
    app()  # pragma: no cover
